@@ -182,7 +182,7 @@ bool ArgumentReader::read(ArgumentVector &args)
     Argument *lastArgInLevel = nullptr;
     vector<const char *> *values = nullptr;
 
-    // iterate through all argument denotations; loop might exit earlier when an denotation is unknown
+    // iterate through all argument denotations; loop might exit earlier when a denotation is unknown
     while (argv != end) {
         // check whether there are still values to read
         if (values && ((lastArgInLevel->requiredValueCount() != Argument::varValueCount) || (lastArgInLevel->flags() & Argument::Flags::Greedy))
@@ -361,6 +361,9 @@ bool ArgumentReader::read(ArgumentVector &args)
             ++parser.m_actualArgc;
             lastArg = lastArgInLevel = matchingArg;
             argDenotation = nullptr;
+            if ((values->size() < matchingArg->requiredValueCount()) && (matchingArg->flags() & Argument::Flags::Greedy)) {
+                continue;
+            }
             read(lastArg->m_subArgs);
             argDenotation = nullptr;
             continue;
@@ -981,9 +984,7 @@ void ArgumentParser::readArgs(int argc, const char *const *argv)
     // check for further arguments
     if (!--argc) {
         // no arguments specified -> flag default argument as present if one is assigned
-        if (m_defaultArg) {
-            m_defaultArg->m_occurrences.emplace_back(0);
-        }
+        assumeDefaultArgument();
         return;
     }
 
@@ -1038,6 +1039,43 @@ void ArgumentParser::resetArgs()
     m_actualArgc = 0;
 }
 
+/*!
+ * \brief Assumes the default argument is present (if not already present).
+ * \remarks
+ * - Adds an occurrence to the default argument so this argument is assumed to be present (even though
+ *   it is not actually present).
+ * - This function is automatically called if there are no arguments present at all.
+ * - This function may be called manually if no actually relevant arguments are present, e.g. if only
+ *   global flags are present but no main operation argument. Note that for this concrete example there
+ *   is also ensureDefaultOperation() (which will call assumeDefaultArgument() internally).
+ */
+void ArgumentParser::assumeDefaultArgument()
+{
+    if (m_defaultArg && m_defaultArg->m_occurrences.empty()) {
+        m_defaultArg->m_occurrences.emplace_back(0);
+    }
+}
+
+/*!
+ * \brief Ensures a main operation argument is present.
+ * \remarks
+ * - Assumes the default argument is present if no other main operation argument or the help argument
+ *   is present.
+ * - Requires the default argument to be an operation argument.
+ */
+void ArgumentParser::ensureDefaultOperation()
+{
+    if (!m_defaultArg || m_defaultArg->isPresent() || !m_defaultArg->denotesOperation()) {
+        return;
+    }
+    for (auto *const arg : m_mainArgs) {
+        if (arg->isPresent() && (arg->denotesOperation() || arg == &m_helpArg)) {
+            return;
+        }
+    }
+    assumeDefaultArgument();
+}
+
 /*!
  * \brief Returns the first operation argument specified by the user or nullptr if no operation has been specified.
  * \remarks Only main arguments are considered. See Argument::specifiedOperation() to check sub arguments of a specific
@@ -1164,7 +1202,7 @@ ArgumentCompletionInfo ArgumentParser::determineCompletionInfo(
         completion.lastSpecifiedArgIndex = static_cast<unsigned int>(argc) - 1;
         completion.lastSpecifiedArg = argv + completion.lastSpecifiedArgIndex;
         for (; completion.lastSpecifiedArg >= argv && **completion.lastSpecifiedArg == '\0';
-             --completion.lastSpecifiedArg, --completion.lastSpecifiedArgIndex)
+            --completion.lastSpecifiedArg, --completion.lastSpecifiedArgIndex)
             ;
     }
 
@@ -1320,6 +1358,25 @@ string ArgumentParser::findSuggestions(int argc, const char *const *argv, unsign
     return suggestionStr;
 }
 
+/*!
+ * \brief Returns a copy of \a escaped with escaping characters removed.
+ */
+static std::string unescape(std::string_view escaped)
+{
+    auto unescaped = std::string();
+    auto onEscaping = false;
+    unescaped.reserve(escaped.size());
+    for (const auto c : escaped) {
+        if (!onEscaping && c == '\\') {
+            onEscaping = true;
+        } else {
+            unescaped += c;
+            onEscaping = false;
+        }
+    }
+    return unescaped;
+}
+
 /*!
  * \brief Prints the bash completion for the specified arguments and the specified \a lastPath.
  * \remarks Arguments must have been parsed before with readSpecifiedArgs(). When calling this method, completionMode must
@@ -1505,23 +1562,13 @@ void ArgumentParser::printBashCompletion(int argc, const char *const *argv, unsi
     }
     // -> completions for files and dirs
     // -> if there's already an "opening", determine the dir part and the file part
-    string actualDir, actualFile;
-    bool haveFileOrDirCompletions = false;
+    auto actualDir = std::string(), actualFile = std::string();
+    auto haveFileOrDirCompletions = false;
     if (argc && currentWordIndex == completionInfo.lastSpecifiedArgIndex && opening) {
-        // the "opening" might contain escaped characters which need to be unescaped first (let's hope this covers all possible escapings)
-        string unescapedOpening(opening);
-        findAndReplace<string>(unescapedOpening, "\\ ", " ");
-        findAndReplace<string>(unescapedOpening, "\\,", ",");
-        findAndReplace<string>(unescapedOpening, "\\[", "[");
-        findAndReplace<string>(unescapedOpening, "\\]", "]");
-        findAndReplace<string>(unescapedOpening, "\\!", "!");
-        findAndReplace<string>(unescapedOpening, "\\#", "#");
-        findAndReplace<string>(unescapedOpening, "\\$", "$");
-        findAndReplace<string>(unescapedOpening, "\\'", "'");
-        findAndReplace<string>(unescapedOpening, "\\\"", "\"");
-        findAndReplace<string>(unescapedOpening, "\\\\", "\\");
+        // the "opening" might contain escaped characters which need to be unescaped first
+        const auto unescapedOpening = unescape(opening);
         // determine the "directory" part
-        string dir = directory(unescapedOpening);
+        auto dir = directory(unescapedOpening);
         if (dir.empty()) {
             actualDir = ".";
         } else {
@@ -1534,7 +1581,7 @@ void ArgumentParser::printBashCompletion(int argc, const char *const *argv, unsi
             actualDir = std::move(dir);
         }
         // determine the "file" part
-        string file = fileName(unescapedOpening);
+        auto file = fileName(unescapedOpening);
         if (file[0] == '\"' || file[0] == '\'') {
             file.erase(0, 1);
         }
@@ -1787,7 +1834,7 @@ void ValueConversion::Helper::ArgumentValueConversionError::throwFailure(const s
     throw ParseError(argumentPath.empty()
             ? argsToString("Conversion of top-level value \"", valueToConvert, "\" to type \"", targetTypeName, "\" failed: ", errorMessage)
             : argsToString("Conversion of value \"", valueToConvert, "\" (for argument --", argumentPath.back()->name(), ") to type \"",
-                targetTypeName, "\" failed: ", errorMessage));
+                  targetTypeName, "\" failed: ", errorMessage));
 }
 
 /*!
@@ -1798,7 +1845,7 @@ void ArgumentOccurrence::throwNumberOfValuesNotSufficient(unsigned long valuesTo
     throw ParseError(path.empty()
             ? argsToString("Expected ", valuesToConvert, " top-level values to be present but only ", values.size(), " have been specified.")
             : argsToString("Expected ", valuesToConvert, " values for argument --", path.back()->name(), " to be present but only ", values.size(),
-                " have been specified."));
+                  " have been specified."));
 }
 
 } // namespace CppUtilities
